import { Subject, BehaviorSubject } from 'rxjs';

import { AxiosInstance } from 'axios';

import { FileUtil } from '@ucap/core';
import { StatusCode } from '@ucap/domain-status';

import {
  NativeService,
  UpdateCheckConfig,
  UpdateInfo,
  NativeType,
  WindowState,
  WindowIdle,
  NotificationRequest,
  NotificationType,
  AppInitInfo
} from '@ucap/native';

import { NotificationService } from './notification.service';

export class BrowserNativeService extends NativeService {
  // tslint:disable-next-line: variable-name
  private _notificationService: NotificationService;
  // tslint:disable-next-line: variable-name
  private _idle_checker: WindowIdleChecker;

  constructor(private axios: AxiosInstance) {
    super();

    this._notificationService = new NotificationService();
  }

  platform_nativeType(): Promise<NativeType> {
    return new Promise<any>((resolve, reject) => {
      try {
        resolve(NativeType.Browser);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_networkInfo(): Promise<any> {
    return new Promise<any>((resolve, reject) => {
      try {
        resolve([{ ip: 'Browser', mac: 'browser' }]);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_execute(executableName: string): Promise<number> {
    return new Promise<number>((resolve, reject) => {
      try {
        resolve(-1);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_openDefaultBrowser(
    url: string,
    options?: {
      name?: string;
      features?: string;
      replace?: boolean;
    }
  ): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        const name = !!options && !!options.name ? options.name : null;
        const features =
          !!options && !!options.features ? options.features : null;
        window.open(url, name, features);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_readFromClipboard(): Promise<{
    text?: string;
    rtf?: string;
    html?: string;
    image?: Buffer;
    imageDataUrl?: string;
  }> {
    return new Promise<{
      text?: string;
      rtf?: string;
      html?: string;
      image?: Buffer;
      imageDataUrl?: string;
    }>((resolve, reject) => {
      try {
        navigator.permissions
          .query({ name: 'clipboard-read' as PermissionName })
          .then((result) => {
            if ('granted' === result.state || 'prompt' === result.state) {
              navigator.clipboard
                .readText()
                .then((value) => {
                  resolve({ text: value });
                })
                .catch((reason) => {
                  reject(reason);
                });
            }
          })
          .catch((reason) => {
            reject(reason);
          });
      } catch (error) {
        reject(error);
      }
    });
  }

  file_save(
    buffer: Buffer,
    fileName: string,
    mimeType: string,
    path?: string
  ): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      try {
        FileUtil.save(buffer, fileName, mimeType)
          .then((fn) => {
            resolve(fn);
          })
          .catch((reason) => {
            reject(reason);
          });
      } catch (error) {
        reject(error);
      }
    });
  }
  file_read(path: string): Promise<Buffer> {
    return new Promise<Buffer>((resolve, reject) => {
      try {
        this.axios
          .get(path, { responseType: 'arraybuffer' })
          .then((res) => {
            resolve(Buffer.from(res.data));
          })
          .catch((reason) => reject(reason));
      } catch (error) {
        reject(error);
      }
    });
  }
  file_openFolder(folderPath?: string, make?: boolean): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(false);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_openItem(filePath?: string): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(false);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_path(
    name: 'home' | 'documents' | 'downloads' | 'music' | 'pictures' | 'videos',
    ...appendPaths: string[]
  ): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_selectForOpen(option: {
    title?: string;
    defaultPath?: string;
    filters?: {
      extensions: string[];
      name: string;
    }[];
    properties?: Array<
      | 'openFile'
      | 'openDirectory'
      | 'multiSelections'
      | 'showHiddenFiles'
      | 'createDirectory'
      | 'promptToCreate'
      | 'noResolveAliases'
      | 'treatPackageAsDirectory'
    >;
    message?: string;
  }): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_selectForSave(option: {
    title?: string;
    defaultPath?: string;
    filters?: {
      extensions: string[];
      name: string;
    }[];
    message?: string;
  }): Promise<{
    canceled: boolean;
    filePath: string;
  }> {
    return new Promise<{
      canceled: boolean;
      filePath: string;
    }>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }

  window_state(): Promise<WindowState> {
    return new Promise<WindowState>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }
  window_onState$(): BehaviorSubject<WindowState> {
    return super.window_onState$();
  }
  window_focused(): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(document.hasFocus());
      } catch (error) {
        reject(error);
      }
    });
  }
  window_onFocus$(): BehaviorSubject<boolean> {
    if (!this._window_onFocusSubject) {
      window.addEventListener(
        'focus',
        this._window_onFocus$CallBack.bind(this)
      );
      window.addEventListener('blur', this._window_onFocus$CallBack.bind(this));
      this.window_focused().then((focused) => {
        this._window_onFocusSubject.next(focused);
      });
    }

    return super.window_onFocus$();
  }
  _window_onFocus$CallBack(event: Event) {
    if (!!this._window_onFocusSubject) {
      this._window_onFocusSubject.next(document.hasFocus());
    }
  }
  window_close(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  window_minimize(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  window_maximize(altKey: boolean = false): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  idle_startCheck(limitTime: number): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        if (!!this._idle_onStateSubject) {
          return;
        }

        this._idle_checker = new WindowIdleChecker({
          limitTime,
          onChange: (idle: WindowIdle) => {
            if (!!this._idle_onStateSubject) {
              this._idle_onStateSubject.next(idle);
            }
          }
        });
        this._idle_checker.start();
        super.idle_onState$();

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  idle_onState$(): BehaviorSubject<WindowIdle> {
    return super.idle_onState$();
  }
  idle_stopCheck(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        if (!!this._idle_checker) {
          this._idle_checker.stop();
          this._idle_checker = undefined;
        }
        if (!!this._idle_onStateSubject) {
          this._idle_onStateSubject.complete();
          this._idle_onStateSubject = undefined;
        }

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  idle_changeLimitTime(limitTime: number): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        if (!this._idle_checker) {
          return;
        }
        this._idle_checker.changeLimitTime(limitTime);

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_version(): Promise<string> {
    return new Promise<any>((resolve, reject) => {
      try {
        resolve('');
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postInit(): Promise<AppInitInfo> {
    return new Promise<AppInitInfo>((resolve, reject) => {
      try {
        resolve({});
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postLogin(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        this._notificationService.requestPermission();

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postLogout(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postDestroy(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_changeAutoLaunch(autoLaunch: boolean): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(true);
      } catch (error) {
        reject(error);
      }
    });
  }
  app_showNotify(req: NotificationRequest): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        this._notificationService.notify(req, () => {
          window.focus();

          switch (req.type) {
            case NotificationType.Event:
              if (!!this._chat_onOpenSubject) {
                this._chat_onOpenSubject.next(req.seq);
              }
              break;
            case NotificationType.Message:
              if (!!this._message_onOpenSubject) {
                this._message_onOpenSubject.next(req.seq);
              }
              break;
            default:
              break;
          }
        });

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_closeAllNotify(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_checkForUpdates(currentVersion: string): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_startCheckForUpdate(config: UpdateCheckConfig): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_onUpdate$(): BehaviorSubject<UpdateInfo> {
    return super.app_onUpdate$();
  }
  app_stopCheckForUpdate(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_applyInstantUpdates(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_exit(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        window.close();
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_onLogout$(): Subject<void> {
    return super.app_onLogout$();
  }
  app_onStatus$(): BehaviorSubject<StatusCode> {
    return super.app_onStatus$();
  }
  app_onShowSetting$(): Subject<void> {
    return super.app_onShowSetting$();
  }
  app_onExit$(): Subject<void> {
    return super.app_onExit$();
  }

  chat_openRoom(openUrl: string): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        window.open(openUrl);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  chat_onOpen$(): Subject<string> {
    return super.chat_onOpen$();
  }

  message_onOpen$(): Subject<string> {
    return super.message_onOpen$();
  }
}

interface WindowIdleCheckerConfig {
  limitTime?: number;
  events?: string[];
  onChange?: (idle: WindowIdle) => void;
}

const defaultWindowIdleCheckerConfig: WindowIdleCheckerConfig = {
  limitTime: 5 * 60 * 1000,
  events: ['mousemove', 'keydown', 'mousedown', 'touchstart'],
  onChange: (idle: WindowIdle) => {}
};

abstract class WindowListener {
  // tslint:disable-next-line: variable-name
  protected events: string[];

  constructor() {}

  start() {
    // tslint:disable-next-line: variable-name
    const __this = this;
    this._addEventListerner(this.events, this._onEvent.bind(this));
  }

  stop() {
    this._removeEventListerner(this.events, this._onEvent.bind(this));
  }

  protected abstract onEvent(event: Event): void;

  private _onEvent(event: Event) {
    this.onEvent(event);
  }

  private _addEventListerner(
    events: string[],
    callback: (event: Event) => void
  ) {
    events.forEach((e) => {
      window.addEventListener(e, callback);
    });
  }

  private _removeEventListerner(
    events: string[],
    callback: (event: Event) => void
  ) {
    events.forEach((e) => {
      window.removeEventListener(e, callback);
    });
  }
}

class WindowIdleChecker extends WindowListener {
  // tslint:disable-next-line: variable-name
  private _timerId: any;
  // tslint:disable-next-line: variable-name
  private _status: WindowIdle;
  // tslint:disable-next-line: variable-name
  private _started: boolean;

  constructor(
    // tslint:disable-next-line: variable-name
    private _config: WindowIdleCheckerConfig
  ) {
    super();
    this._config = Object.assign({}, defaultWindowIdleCheckerConfig, _config);
    this._status = WindowIdle.Active;
    this.events = this._config.events;
  }

  start() {
    if (!!this._timerId) {
      clearInterval(this._timerId);
    }

    this._timerId = setInterval(() => {
      if (WindowIdle.Active === this._status) {
        this._status = WindowIdle.Idle;
        this._config.onChange(this._status);
      }
    }, this._config.limitTime);

    if (!this._started) {
      this._started = true;
    }

    super.start();
  }

  stop() {
    if (!!this._timerId) {
      clearInterval(this._timerId);
      this._timerId = undefined;
    }

    if (!!this._started) {
      super.stop();
    }
  }

  changeLimitTime(limitTime: number) {
    this._config = {
      ...this._config,
      limitTime
    };
    this.start();
  }

  protected onEvent(event: Event): void {
    if (WindowIdle.Idle === this._status) {
      this._status = WindowIdle.Active;
      this._config.onChange(this._status);
    }
  }
}
