import moment from 'moment';

import { Selector, createSelector } from '@ngrx/store';
import { EntityState, createEntityAdapter } from '@ngrx/entity';

import {
  RoomInfo,
  UserInfo as RoomUserInfo,
  UserInfoShort as RoomUserInfoShort
} from '@ucap/domain-chat';

export interface RoomState extends EntityState<RoomInfo> {
  syncDate: string;
}
export const adapterRoom = createEntityAdapter<RoomInfo>({
  selectId: (roomInfo) => roomInfo.roomId,
  sortComparer: (a, b) => {
    return (
      moment(b.finalEventDate).toDate().getTime() -
      moment(a.finalEventDate).toDate().getTime()
    );
  }
});

export interface RoomUserMap {
  roomId: string;
  userInfos: RoomUserInfo[];
}

export interface RoomUserShortMap {
  roomId: string;
  userInfos: RoomUserInfoShort[];
}

export interface RoomUserState extends EntityState<RoomUserMap> {}
export const adapterRoomUser = createEntityAdapter<RoomUserMap>({
  selectId: (roomUserDetailData) => roomUserDetailData.roomId
});

export interface RoomUserShortState extends EntityState<RoomUserShortMap> {}
export const adapterRoomUserShort = createEntityAdapter<RoomUserShortMap>({
  selectId: (roomUserShortData) => roomUserShortData.roomId
});

export interface State {
  rooms: RoomState;
  roomUsers: RoomUserState;
  roomUsersShort: RoomUserShortState;
  standbyRooms: string[];
}

const roomInitialState: RoomState = adapterRoom.getInitialState({
  syncDate: ''
});
const roomUserInitialState: RoomUserState = adapterRoomUser.getInitialState({});
const roomUserShortInitialState: RoomUserShortState = adapterRoomUserShort.getInitialState(
  {}
);

export const initialState: State = {
  rooms: roomInitialState,
  roomUsers: roomUserInitialState,
  roomUsersShort: roomUserShortInitialState,
  standbyRooms: []
};

const {
  selectAll: selectAllForRoom,
  selectEntities: selectEntitiesForRoom,
  selectIds: selectIdsForRoom,
  selectTotal: selectTotalForRoom
} = adapterRoom.getSelectors();

const {
  selectAll: selectAllForRoomUser,
  selectEntities: selectEntitiesForRoomUser,
  selectIds: selectIdsForRoomUser,
  selectTotal: selectTotalForRoomUser
} = adapterRoomUser.getSelectors();

const {
  selectAll: selectAllForRoomUserShort,
  selectEntities: selectEntitiesForRoomUserShort,
  selectIds: selectIdsForRoomUserShort,
  selectTotal: selectTotalForRoomUserShort
} = adapterRoomUserShort.getSelectors();

export function selectors<S>(selector: Selector<any, State>) {
  const selectRooms = createSelector(selector, (state: State) => state.rooms);

  const selectRoomUsers = createSelector(
    selector,
    (state: State) => state.roomUsers
  );

  const selectRoomUsersShort = createSelector(
    selector,
    (state: State) => state.roomUsersShort
  );

  return {
    rooms: createSelector(selectRooms, selectAllForRoom),
    room: createSelector(
      selectRooms,
      (roomState: RoomState, roomId: string) => {
        return roomState.entities && roomState.entities[roomId];
      }
    ),
    roomsSyncDate: createSelector(
      selectRooms,
      (roomState) => roomState.syncDate
    ),
    roomUsers: createSelector(selectRoomUsers, selectAllForRoomUser),
    roomUser: createSelector(
      selectRoomUsers,
      (roomUserState: RoomUserState, roomId: string) => {
        return roomUserState.entities && roomUserState.entities[roomId];
      }
    ),
    roomUsersShort: createSelector(
      selectRoomUsersShort,
      selectAllForRoomUserShort
    ),
    roomUserShort: createSelector(
      selectRoomUsersShort,
      (roomUserShortState: RoomUserShortState, roomId: string) => {
        return (
          roomUserShortState.entities && roomUserShortState.entities[roomId]
        );
      }
    ),
    standbyRooms: createSelector(
      selector,
      (state: State) => state.standbyRooms
    ),
    unreadTotal: createSelector(selectRooms, (roomState: RoomState) => {
      let unreadTotal = 0;

      // tslint:disable-next-line: forin
      for (const key in roomState.ids) {
        const roomId = roomState.ids[key];
        if (roomId === undefined) {
          continue;
        }

        if (roomState.entities.hasOwnProperty(roomId)) {
          const room = roomState.entities[roomId];
          unreadTotal += room.noReadCnt;
        }
      }

      return unreadTotal;
    })
  };
}
