import {
  Component,
  OnInit,
  EventEmitter,
  Output,
  ViewChild,
  ElementRef,
  OnDestroy,
  ChangeDetectionStrategy,
  ChangeDetectorRef
} from '@angular/core';

import { Subject, Subscription } from 'rxjs';
import { debounceTime, takeUntil } from 'rxjs/operators';

import { MatDialog } from '@angular/material/dialog';

import { I18nService } from '@ucap/ng-i18n';

import {
  AlertDialogComponent,
  AlertDialogData,
  AlertDialogResult
} from '@ucap/ng-ui/core';
import { PhoneNumberPipe } from '@ucap/ng-ui/phone';

@Component({
  selector: 'ucap-call-dialpad',
  templateUrl: './dialpad.component.html',
  styleUrls: ['./dialpad.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush
})
export class DialpadComponent implements OnInit, OnDestroy {
  @Output()
  sendCall = new EventEmitter<string>();

  @ViewChild('phoneInput', { static: false })
  phoneInput: ElementRef<HTMLInputElement>;

  inputSubscription: Subscription;
  isShowRemoveBtn: boolean;

  private subject: Subject<string> = new Subject();
  private ngOnDestroySubject: Subject<void> = new Subject();
  constructor(
    private dialog: MatDialog,
    private i18nService: I18nService,
    private changeDetectorRef: ChangeDetectorRef
  ) {}

  ngOnInit() {
    this.isShowRemoveBtn = false;

    this.inputSubscription = this.subject
      .pipe(takeUntil(this.ngOnDestroySubject), debounceTime(400))
      .subscribe((num) => {
        this._makePhoneNumber(num);
      });
  }

  ngOnDestroy(): void {
    if (!!this.ngOnDestroySubject) {
      this.ngOnDestroySubject.next();
      this.ngOnDestroySubject.complete();
    }
  }

  /** About Input field */
  onBlur(event): void {
    this.phoneInput.nativeElement.focus();
  }
  onKeyup(val): void {
    this.subject.next(val.replace(/[^#*\d]/g, ''));
  }
  onClickAllRemove(): void {
    this.phoneInput.nativeElement.value = '';
    this.isShowRemoveBtn = false;
  }
  onClickRemove(): void {
    let curNumberStr = this.phoneInput.nativeElement.value;
    curNumberStr = curNumberStr.substring(0, curNumberStr.length - 1);
    this._makePhoneNumber(curNumberStr);
  }

  /** About Dialpad */
  onClickNumber(num: string): void {
    this._makePhoneNumber(this.phoneInput.nativeElement.value.concat(num));
  }

  onClickCall(): void {
    const curNumberStr = this.phoneInput.nativeElement.value;
    const phoneNum = curNumberStr.replace(/[^#*\d]/g, '');

    if (!!phoneNum && phoneNum.length > 0) {
      this.sendCall.emit(phoneNum);
    } else {
      this.dialog.open<
        AlertDialogComponent,
        AlertDialogData,
        AlertDialogResult
      >(AlertDialogComponent, {
        panelClass: 'min-create-dialog',
        data: {
          title: this.i18nService.t('call:label.error'),
          message: this.i18nService.t('call:errors.inputCallNumber')
        }
      });
      return;
    }
  }

  private _makePhoneNumber(numberStr: string) {
    if (numberStr.trim() === '') {
      this.phoneInput.nativeElement.value = '';
      this.isShowRemoveBtn = false;
      this.changeDetectorRef.markForCheck();
      return;
    }

    this.isShowRemoveBtn = true;
    if (this._getSpecialSymbol(numberStr)) {
      this.phoneInput.nativeElement.value = numberStr;
    } else {
      this.phoneInput.nativeElement.value = `${new PhoneNumberPipe().transform(
        numberStr
      )}`;
    }
    this.changeDetectorRef.markForCheck();
  }

  private _getSpecialSymbol(num: string): boolean {
    const curNumberStr = this.phoneInput.nativeElement.value;

    const asteriskIndex = curNumberStr.indexOf('*');
    const hashIndex = curNumberStr.indexOf('#');

    if (
      asteriskIndex > -1 ||
      hashIndex > -1 ||
      0 === num.localeCompare('*') ||
      0 === num.localeCompare('#')
    ) {
      return true;
    } else {
      return false;
    }
  }
}
