import { Subject } from 'rxjs';
import { takeUntil } from 'rxjs/operators';

import {
  Component,
  OnInit,
  OnDestroy,
  Input,
  ViewChild,
  ContentChild,
  TemplateRef,
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Directive,
  AfterViewInit
} from '@angular/core';

import { FlatTreeControl } from '@angular/cdk/tree';

import { MatTreeFlattener, MatTree } from '@angular/material/tree';

import { PerfectScrollbarDirective } from 'ngx-perfect-scrollbar';

import {
  VirtualScrollTreeFlatDataSource,
  VirtualScrollViewportComponent
} from '@ucap/ng-ui/scrolling';
import { CallHistory } from '@ucap/domain-call';

export interface HistoryNode {
  nodeType: string;
  historyInfo?: CallHistory;
  children?: HistoryNode[];
}

export interface FlatNode {
  expandable: boolean;
  level: number;
  node: HistoryNode;
}

@Directive({
  selector: '[ucapCallHistoryExpansionNode]'
})
export class HistoryExpansionNodeDirective {}

@Directive({
  selector: '[ucapCallHistoryExpansionHeader]'
})
export class HistoryExpansionHeaderDirective {}

@Component({
  selector: 'ucap-call-history-expansion',
  templateUrl: './history-expansion.component.html',
  styleUrls: ['./history-expansion.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush
})
export class HistoryExpansionComponent
  implements OnInit, OnDestroy, AfterViewInit {
  @Input()
  set historyGroup(list: { division: string; historyList: CallHistory[] }[]) {
    if (!list || 0 === list.length) {
    } else {
      list.sort((a, b) =>
        a.division < b.division ? 1 : a.division > b.division ? -1 : 0
      );

      for (const item of list) {
        const nodeType = item.division;
        const node: HistoryNode = {
          nodeType,
          children: []
        };

        item.historyList.sort((a, b) =>
          a.callLogSeq < b.callLogSeq ? 1 : a.callLogSeq > b.callLogSeq ? -1 : 0
        );

        item.historyList.forEach((historyInfo) => {
          node.children.push({
            nodeType,
            historyInfo
          });
        });

        if (!!this.nodeMap.get(item.division)) {
          this.nodeMap.get(item.division)[0].children = node.children;
        } else {
          this.nodeMap.set(item.division, [node]);
        }
      }

      // deleted filtering.
      this.nodeMap = new Map(
        [...this.nodeMap.entries()].filter(
          (map) => list.findIndex((item) => item.division === map[0]) > -1
        )
      );

      // sorting.
      this.nodeMap = new Map(
        [...this.nodeMap.entries()].sort((a, b) =>
          b[0] > a[0] ? 1 : b[0] < a[0] ? -1 : 0
        )
      );
    }
    this.refreshNodes();
  }

  @Input()
  vsItemSize: number;

  @ViewChild('treeList', { static: false })
  treeList: MatTree<FlatNode>;

  @ViewChild('cvsvList', { static: false })
  cvsvList: VirtualScrollViewportComponent;

  @ViewChild(PerfectScrollbarDirective, { static: false })
  psDirectiveRef?: PerfectScrollbarDirective;

  @ContentChild(HistoryExpansionNodeDirective, {
    read: TemplateRef,
    static: false
  })
  nodeTemplate: TemplateRef<HistoryExpansionNodeDirective>;

  @ContentChild(HistoryExpansionHeaderDirective, {
    read: TemplateRef,
    static: false
  })
  headerTemplate: TemplateRef<HistoryExpansionHeaderDirective>;

  treeControl: FlatTreeControl<FlatNode>;
  treeFlattener: MatTreeFlattener<HistoryNode, FlatNode>;
  dataSource: VirtualScrollTreeFlatDataSource<HistoryNode, FlatNode>;

  private nodeMap: Map<string, HistoryNode[]> = new Map();
  // tslint:disable-next-line: variable-name
  private _ngOnDestroySubject: Subject<void> = new Subject();

  constructor(private changeDetectorRef: ChangeDetectorRef) {
    this.treeControl = new FlatTreeControl<FlatNode>(
      (node) => node.level,
      (node) => node.expandable
    );

    this.treeFlattener = new MatTreeFlattener<HistoryNode, FlatNode>(
      (node: HistoryNode, level: number) => {
        return {
          expandable: !!node.children && node.children.length > 0,
          level,
          nodeType: node.nodeType,
          node
        };
      },
      (node) => node.level,
      (node) => node.expandable,
      (node) => node.children
    );

    this.dataSource = new VirtualScrollTreeFlatDataSource<
      HistoryNode,
      FlatNode
    >(this.treeControl, this.treeFlattener);
  }

  ngOnInit(): void {
    this.dataSource.expandedData$
      .pipe(takeUntil(this._ngOnDestroySubject))
      .subscribe((datas) => {
        if (!!this.psDirectiveRef) {
          this.psDirectiveRef.update();
        }
      });
  }

  ngOnDestroy(): void {
    if (!!this._ngOnDestroySubject) {
      this._ngOnDestroySubject.next();
      this._ngOnDestroySubject.complete();
    }
  }

  ngAfterViewInit(): void {
    this.dataSource.virtualScrollViewport = this.cvsvList;
  }

  isHeader = (_: number, node: FlatNode) => 0 === node.level;

  private refreshNodes() {
    const rootNode: HistoryNode[] = [];
    this.nodeMap.forEach((node) => rootNode.push(...node));

    this.dataSource.data = rootNode;
    this.treeControl.expandAll();
    this.changeDetectorRef.detectChanges();
  }
}
