import { Subject } from 'rxjs';
import { takeUntil } from 'rxjs/operators';

import {
  Component,
  OnInit,
  OnDestroy,
  Input,
  ViewChild,
  ContentChild,
  TemplateRef,
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Directive,
  AfterViewInit
} from '@angular/core';

import { FlatTreeControl } from '@angular/cdk/tree';

import { MatTreeFlattener, MatTree } from '@angular/material/tree';

import { PerfectScrollbarDirective } from 'ngx-perfect-scrollbar';

import { RoomInfo } from '@ucap/domain-chat';

import {
  VirtualScrollTreeFlatDataSource,
  VirtualScrollViewportComponent
} from '@ucap/ng-ui/scrolling';

export interface RoomNode {
  nodeType: string;
  roomInfo?: RoomInfo;
  children?: RoomNode[];
}

export interface FlatNode {
  expandable: boolean;
  level: number;
  node: RoomNode;
}

@Directive({
  selector: '[ucapChatRoomExpansionNode]'
})
export class RoomExpansionNodeDirective {}

@Directive({
  selector: '[ucapChatRoomExpansionHeader]'
})
export class RoomExpansionHeaderDirective {}

@Component({
  selector: 'ucap-chat-room-expansion',
  templateUrl: './room-expansion.component.html',
  styleUrls: ['./room-expansion.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush
})
export class RoomExpansionComponent
  implements OnInit, OnDestroy, AfterViewInit {
  @Input()
  set roomGroup(list: { division: string; roomList: RoomInfo[] }[]) {
    if (!list || 0 === list.length) {
    } else {
      list.sort((a, b) =>
        a.division < b.division ? 1 : a.division > b.division ? -1 : 0
      );

      for (const item of list) {
        const nodeType = item.division;
        const node: RoomNode = {
          nodeType,
          children: []
        };

        item.roomList.sort((a, b) =>
          a.finalEventDate < b.finalEventDate
            ? 1
            : a.finalEventDate > b.finalEventDate
            ? -1
            : 0
        );

        item.roomList.forEach((roomInfo) => {
          node.children.push({
            nodeType,
            roomInfo
          });
        });

        if (!!this.nodeMap.get(item.division)) {
          this.nodeMap.get(item.division)[0].children = node.children;
        } else {
          this.nodeMap.set(item.division, [node]);
        }
      }

      // deleted filtering.
      this.nodeMap = new Map(
        [...this.nodeMap.entries()].filter(
          (map) => list.findIndex((item) => item.division === map[0]) > -1
        )
      );

      // sorting.
      this.nodeMap = new Map(
        [...this.nodeMap.entries()].sort((a, b) =>
          b[0] > a[0] ? 1 : b[0] < a[0] ? -1 : 0
        )
      );
    }
    this.refreshNodes();
  }

  @Input()
  vsItemSize: number;

  @ViewChild('treeList', { static: false })
  treeList: MatTree<FlatNode>;

  @ViewChild('cvsvList', { static: false })
  cvsvList: VirtualScrollViewportComponent;

  @ViewChild(PerfectScrollbarDirective, { static: false })
  psDirectiveRef?: PerfectScrollbarDirective;

  @ContentChild(RoomExpansionNodeDirective, {
    read: TemplateRef,
    static: false
  })
  nodeTemplate: TemplateRef<RoomExpansionNodeDirective>;

  @ContentChild(RoomExpansionHeaderDirective, {
    read: TemplateRef,
    static: false
  })
  headerTemplate: TemplateRef<RoomExpansionHeaderDirective>;

  treeControl: FlatTreeControl<FlatNode>;
  treeFlattener: MatTreeFlattener<RoomNode, FlatNode>;
  dataSource: VirtualScrollTreeFlatDataSource<RoomNode, FlatNode>;

  private nodeMap: Map<string, RoomNode[]> = new Map();
  // tslint:disable-next-line: variable-name
  private _ngOnDestroySubject: Subject<void> = new Subject();

  constructor(private changeDetectorRef: ChangeDetectorRef) {
    this.treeControl = new FlatTreeControl<FlatNode>(
      (node) => node.level,
      (node) => node.expandable
    );

    this.treeFlattener = new MatTreeFlattener<RoomNode, FlatNode>(
      (node: RoomNode, level: number) => {
        return {
          expandable: !!node.children && node.children.length > 0,
          level,
          nodeType: node.nodeType,
          node
        };
      },
      (node) => node.level,
      (node) => node.expandable,
      (node) => node.children
    );

    this.dataSource = new VirtualScrollTreeFlatDataSource<RoomNode, FlatNode>(
      this.treeControl,
      this.treeFlattener
    );
  }

  ngOnInit(): void {
    this.dataSource.expandedData$
      .pipe(takeUntil(this._ngOnDestroySubject))
      .subscribe((datas) => {
        if (!!this.psDirectiveRef) {
          this.psDirectiveRef.update();
        }
      });
  }

  ngOnDestroy(): void {
    if (!!this._ngOnDestroySubject) {
      this._ngOnDestroySubject.next();
      this._ngOnDestroySubject.complete();
    }
  }

  ngAfterViewInit(): void {
    this.dataSource.virtualScrollViewport = this.cvsvList;
  }

  isHeader = (_: number, node: FlatNode) => 0 === node.level;

  private refreshNodes() {
    const rootNode: RoomNode[] = [];
    this.nodeMap.forEach((node) => rootNode.push(...node));

    this.dataSource.data = rootNode;
    this.treeControl.expandAll();
    this.changeDetectorRef.detectChanges();
  }
}
