import { Subject } from 'rxjs';
import { takeUntil } from 'rxjs/operators';

import {
  Component,
  OnInit,
  OnDestroy,
  Input,
  Output,
  EventEmitter,
  ViewChild,
  ContentChild,
  TemplateRef,
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Directive,
  AfterViewInit
} from '@angular/core';

import { FlatTreeControl } from '@angular/cdk/tree';

import { MatTreeFlattener, MatTree } from '@angular/material/tree';

import { UserInfoSS, UserInfoF, UserInfoDN } from '@ucap/domain-organization';
import { GroupInfoDetail } from '@ucap/domain-group';

import {
  VirtualScrollTreeFlatDataSource,
  VirtualScrollViewportComponent
} from '@ucap/ng-ui/scrolling';

import { PerfectScrollbarDirective } from 'ngx-perfect-scrollbar';

export enum NodeType {
  None = 'None',
  Profile = 'Profile',
  Favorite = 'Favorite',
  Buddy = 'Buddy',
  Default = 'Default'
}

export interface GroupNode {
  nodeType: NodeType;
  userInfo?: UserInfoF;
  groupDetail?: GroupInfoDetail;
  children?: GroupNode[];
}

export interface FlatNode {
  expandable: boolean;
  level: number;
  node: GroupNode;
}

@Directive({
  selector: '[ucapGroupExpansionNode]'
})
export class ExpansionNodeDirective {}

@Directive({
  selector: '[ucapGroupExpansionFavoriteHeader]'
})
export class ExpansionFavoriteHeaderDirective {}

@Directive({
  selector: '[ucapGroupExpansionDefaultHeader]'
})
export class ExpansionDefaultHeaderDirective {}

@Directive({
  selector: '[ucapGroupExpansionBuddyHeader]'
})
export class ExpansionBuddyHeaderDirective {}

@Component({
  selector: 'ucap-group-expansion',
  templateUrl: './expansion.component.html',
  styleUrls: ['./expansion.component.scss'],
  changeDetection: ChangeDetectionStrategy.OnPush
})
export class ExpansionComponent implements OnInit, OnDestroy, AfterViewInit {
  @Input()
  displayOrder: NodeType[] = [];

  @Input()
  vsItemSize: number;

  @Input()
  cacheSize = 0;

  @Input()
  set profile(userInfo: UserInfoF) {
    if (!userInfo) {
      this.nodeMap.set(NodeType.Profile, []);
    } else {
      const node: GroupNode = {
        nodeType: NodeType.Profile,
        userInfo,
        children: []
      };

      this.nodeMap.set(NodeType.Profile, [node]);
    }

    this.refreshNodes();
  }

  @Input()
  set favorites(userInfos: UserInfoF[]) {
    if (!userInfos || 0 === userInfos.length) {
      this.nodeMap.set(NodeType.Favorite, []);
    } else {
      const node: GroupNode = {
        nodeType: NodeType.Favorite,
        groupDetail: {
          seq: -9999,
          name: NodeType.Favorite,
          isActive: true,
          userSeqs: userInfos.map((userInfo) => String(userInfo.seq))
        } as GroupInfoDetail,
        children: []
      };

      userInfos.forEach((userInfo) => {
        node.children.push({
          nodeType: NodeType.Profile,
          groupDetail: node.groupDetail,
          userInfo
        });
      });

      this.nodeMap.set(NodeType.Favorite, [node]);
    }

    this.refreshNodes();
  }

  @Input()
  set groupBuddies(list: { group: GroupInfoDetail; buddyList: UserInfoF[] }[]) {
    if (!list || 0 === list.length) {
      this.nodeMap.set(NodeType.Default, []);
      this.nodeMap.set(NodeType.Buddy, []);
    } else {
      this.nodeMap.set(NodeType.Default, []);
      this.nodeMap.set(NodeType.Buddy, []);

      this.groupList = list;

      for (const item of list) {
        let nodeType = NodeType.Buddy;
        if (0 === item.group.seq) {
          nodeType = NodeType.Default;
        }

        const node: GroupNode = {
          nodeType,
          groupDetail: item.group,
          children: []
        };

        item.buddyList.sort((a, b) =>
          a.order < b.order
            ? -1
            : a.order > b.order
            ? 1
            : a.name < b.name
            ? -1
            : a.name > b.name
            ? 1
            : 0
        );

        item.buddyList.forEach((userInfo) => {
          node.children.push({
            nodeType: NodeType.Profile,
            groupDetail: item.group,
            userInfo
          });
        });

        if (NodeType.Buddy === nodeType) {
          this.nodeMap.get(NodeType.Buddy).push(node);
        } else {
          this.nodeMap.get(NodeType.Default).push(node);
        }
      }
    }

    this.refreshNodes();
  }

  @Input()
  selectedGroupHeader: GroupInfoDetail;

  @Input()
  checkable = false;

  @Input()
  moreable = false;

  @Input()
  selectedUserList?: (UserInfoF | UserInfoSS | UserInfoF | UserInfoDN)[] = [];

  @Input()
  groupOpenInfo: { groupSeqs: number[]; lastGroupSeq: number };

  @Input()
  unselectableUserList?: (
    | UserInfoF
    | UserInfoSS
    | UserInfoF
    | UserInfoDN
  )[] = [];

  @Output()
  clickMoreMenu = new EventEmitter<{
    event: MouseEvent;
    node: FlatNode;
  }>();

  @Output()
  checkGroup = new EventEmitter<{
    isChecked: boolean;
    groupBuddyList: { group: GroupInfoDetail; buddyList: UserInfoF[] };
  }>();

  @Output()
  clickNode = new EventEmitter<{ node: FlatNode; isExpand: boolean }>();

  @ViewChild('treeList', { static: false })
  treeList: MatTree<FlatNode>;

  @ViewChild('cvsvList', { static: false })
  cvsvList: VirtualScrollViewportComponent;

  @ViewChild(PerfectScrollbarDirective, { static: false })
  psDirectiveRef: PerfectScrollbarDirective;

  @ContentChild(ExpansionNodeDirective, {
    read: TemplateRef,
    static: false
  })
  nodeTemplate: TemplateRef<ExpansionNodeDirective>;

  @ContentChild(ExpansionFavoriteHeaderDirective, {
    read: TemplateRef,
    static: false
  })
  favoriteHeaderTemplate: TemplateRef<ExpansionFavoriteHeaderDirective>;

  @ContentChild(ExpansionDefaultHeaderDirective, {
    read: TemplateRef,
    static: false
  })
  defaultHeaderTemplate: TemplateRef<ExpansionDefaultHeaderDirective>;

  @ContentChild(ExpansionBuddyHeaderDirective, {
    read: TemplateRef,
    static: false
  })
  buddyHeaderTemplate: TemplateRef<ExpansionBuddyHeaderDirective>;

  treeControl: FlatTreeControl<FlatNode>;
  treeFlattener: MatTreeFlattener<GroupNode, FlatNode>;
  dataSource: VirtualScrollTreeFlatDataSource<GroupNode, FlatNode>;

  NodeType = NodeType;
  groupList: { group: GroupInfoDetail; buddyList: UserInfoF[] }[];

  private nodeMap: Map<NodeType, GroupNode[]> = new Map();
  // tslint:disable-next-line: variable-name
  private _ngOnDestroySubject: Subject<void> = new Subject();

  constructor(private changeDetectorRef: ChangeDetectorRef) {
    this.treeControl = new FlatTreeControl<FlatNode>(
      (node) => node.level,
      (node) => node.expandable
    );

    this.treeFlattener = new MatTreeFlattener<GroupNode, FlatNode>(
      (node: GroupNode, level: number) => {
        return {
          expandable: !!node.children && node.children.length > 0,
          level,
          nodeType: node.nodeType,
          node
        };
      },
      (node) => node.level,
      (node) => node.expandable,
      (node) => node.children
    );

    this.dataSource = new VirtualScrollTreeFlatDataSource<GroupNode, FlatNode>(
      this.treeControl,
      this.treeFlattener
    );
  }

  ngOnInit(): void {
    this.dataSource.expandedData$
      .pipe(takeUntil(this._ngOnDestroySubject))
      .subscribe((datas) => {
        if (!!this.psDirectiveRef) {
          this.psDirectiveRef.update();
        }
      });
  }

  ngOnDestroy(): void {
    if (!!this._ngOnDestroySubject) {
      this._ngOnDestroySubject.next();
      this._ngOnDestroySubject.complete();
    }
  }

  ngAfterViewInit(): void {
    this.dataSource.virtualScrollViewport = this.cvsvList;
  }

  // trackBy = (_: number, node: FlatNode) => node.node.groupDetail.seq;

  onClickHeaderMenu(event: MouseEvent, node: FlatNode) {
    this.clickMoreMenu.emit({ event, node });
  }

  isCheckedGroup(node: FlatNode): boolean {
    const groupDetail = node.node.groupDetail;

    if (!groupDetail || groupDetail === undefined) {
      return false;
    }

    if (groupDetail.userSeqs.length === 0) {
      return false;
    }

    if (!!this.selectedUserList && this.selectedUserList.length > 0) {
      let allExist = true;
      groupDetail.userSeqs.some((seq) => {
        if (
          this.selectedUserList.filter(
            (item) => String(item.seq) === String(seq)
          ).length === 0
        ) {
          allExist = false;
          return true;
        }
      });
      return allExist;
    }
    return false;
  }

  isCheckableGroup(node: FlatNode): boolean {
    if (!!this.unselectableUserList && this.unselectableUserList.length > 0) {
      const groupDetail = node.node.groupDetail;
      let allExist = true;
      groupDetail.userSeqs.some((seq) => {
        if (
          this.unselectableUserList.filter((item) => item.seq === seq)
            .length === 0
        ) {
          allExist = false;
          return true;
        }
      });

      if (allExist) {
        return false;
      }
    }

    return true;
  }

  onChangeCheckGroup(value: boolean, node: FlatNode) {
    const trgGroup = node.node.groupDetail;

    const groupInfos = this.groupList.filter(
      (groupInfo) => groupInfo.group.seq === trgGroup.seq
    );

    if (groupInfos.length > 0) {
      this.checkGroup.emit({
        isChecked: value,
        groupBuddyList: groupInfos[0]
      });
    } else if (node.node.nodeType === NodeType.Favorite) {
      this.checkGroup.emit({
        isChecked: value,
        groupBuddyList: {
          group: trgGroup,
          buddyList: node.node.children.map((child) => child.userInfo)
        }
      });
    }
  }

  isHeader = (_: number, node: FlatNode) =>
    NodeType.Profile !== node.node.nodeType && 0 === node.level;

  expandMore() {
    // this.treeList.treeControl.expandAll();
    this.treeControl.expandAll();
    if (!!this.psDirectiveRef) {
      this.psDirectiveRef.scrollToTop();
    }
  }

  expandLess() {
    this.treeList.treeControl.collapseAll();
    if (!!this.psDirectiveRef) {
      this.psDirectiveRef.scrollToTop();
    }
  }

  onClickNode(event: Event, node: FlatNode) {
    event.stopPropagation();

    this.clickNode.emit({ node, isExpand: this.treeControl.isExpanded(node) });
  }

  expand(params: { groupSeqs: number[]; lastGroupSeq: number }) {
    if (
      !this.treeControl.dataNodes ||
      !params.groupSeqs ||
      0 === params.groupSeqs.length
    ) {
      return;
    }
    this.groupOpenInfo = params;

    const flatNodes: FlatNode[] = [];
    params.groupSeqs.forEach((s) => {
      let node: FlatNode;

      this.treeControl.dataNodes.every((n) => {
        if (
          n.node.nodeType !== NodeType.Profile &&
          n.node.groupDetail.seq === s
        ) {
          node = n;
          return false;
        }
        return true;
      });
      if (!!node) {
        flatNodes.push(node);
      }
    });
    if (0 < flatNodes.length) {
      this.treeControl.expansionModel.select(...flatNodes);
    }

    setTimeout(
      () => {
        const curIndex = this.dataSource.expandedDataSubject.value.findIndex(
          (flatNode) => {
            if (
              flatNode.node.nodeType !== NodeType.Profile &&
              flatNode.node.nodeType !== NodeType.None &&
              flatNode.node.groupDetail.seq === params.lastGroupSeq
            ) {
              return flatNode.node.groupDetail.seq === params.lastGroupSeq;
            }
          }
        );
        if (-1 < curIndex) {
          if ('vertical' === this.cvsvList.orientation) {
            this.psDirectiveRef.scrollToY(
              this.cvsvList.offsetForIndex(curIndex)
            );
          }
        }
      },
      0 < flatNodes.length &&
        0 === this.dataSource.expandedDataSubject.value.length
        ? 100
        : 0
    );
  }

  private refreshNodes() {
    const rootNode: GroupNode[] = [];

    for (const nodeType of this.displayOrder) {
      const subNode = this.nodeMap.get(nodeType);
      if (!!subNode && 0 < subNode.length) {
        rootNode.push(...subNode);
      }
    }

    this.dataSource.data = rootNode;
    this.changeDetectorRef.markForCheck();

    if (!!this.groupOpenInfo && this.groupOpenInfo.groupSeqs.length > 0) {
      this.expand(this.groupOpenInfo);
    }
  }
}
