import { takeUntil, startWith } from 'rxjs/operators';

import {
  Component,
  OnDestroy,
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Input,
  ElementRef,
  ViewChild,
  Directive,
  ContentChildren,
  QueryList,
  Optional,
  Inject,
  Attribute,
  NgZone,
  AfterContentChecked,
  AfterContentInit,
  forwardRef,
  HostBinding
} from '@angular/core';
import { ANIMATION_MODULE_TYPE } from '@angular/platform-browser/animations';

import { FocusableOption, FocusMonitor } from '@angular/cdk/a11y';
import { BooleanInput, coerceBooleanProperty } from '@angular/cdk/coercion';
import { Platform } from '@angular/cdk/platform';
import { Directionality } from '@angular/cdk/bidi';
import { ViewportRuler } from '@angular/cdk/overlay';

import {
  CanDisable,
  HasTabIndex,
  HasTabIndexCtor,
  CanDisableCtor,
  mixinDisabled,
  mixinTabIndex,
  ThemePalette
} from '@angular/material/core';

import {
  PaginatedHeaderDirective,
  PaginatedHeaderItem
} from '../directives/paginated-header';

class BreadcrumbNodeMixinBase {}
// tslint:disable-next-line: variable-name
const _BreadcrumbNodeMixinBase: HasTabIndexCtor &
  CanDisableCtor &
  typeof BreadcrumbNodeMixinBase = mixinTabIndex(
  mixinDisabled(BreadcrumbNodeMixinBase)
);

@Directive()
export abstract class _BreadcrumbNodeBase extends _BreadcrumbNodeMixinBase
  implements OnDestroy, CanDisable, HasTabIndex, FocusableOption {
  static ngAcceptInputType_disabled: BooleanInput;

  protected _isActive = false;

  @Input()
  get active(): boolean {
    return this._isActive;
  }
  set active(value: boolean) {
    if (value !== this._isActive) {
      this._isActive = value;
      this._breadcrumb.updateActiveLink(this.elementRef);
    }
  }

  constructor(
    private _breadcrumb: _BreadcrumbBase,
    public elementRef: ElementRef,
    @Attribute('tabindex') tabIndex: string,
    private _focusMonitor: FocusMonitor
  ) {
    super();

    this.tabIndex = parseInt(tabIndex, 10) || 0;

    _focusMonitor.monitor(elementRef);
  }

  focus() {
    this.elementRef.nativeElement.focus();
  }

  ngOnDestroy() {
    this._focusMonitor.stopMonitoring(this.elementRef);
  }
}

@Directive({
  selector: '[ucap-breadcrumb-node], [ucapBreadcrumbNode]',
  exportAs: 'ucapBreadcrumbNode',
  inputs: ['disabled', 'tabIndex'],
  host: {
    class: 'ucap-breadcrumb-node mat-focus-indicator',
    '[attr.aria-current]': 'active ? "page" : null',
    '[attr.aria-disabled]': 'disabled',
    '[attr.tabIndex]': 'tabIndex',
    '[class.ucap-breadcrumb-disabled]': 'disabled',
    '[class.ucap-breadcrumb-label-active]': 'active'
  }
})
export class BreadcrumbNodeDirective extends _BreadcrumbNodeBase
  implements OnDestroy {
  constructor(
    breadcrumb: BreadcrumbComponent,
    elementRef: ElementRef,
    @Attribute('tabindex') tabIndex: string,
    focusMonitor: FocusMonitor
  ) {
    super(breadcrumb, elementRef, tabIndex, focusMonitor);
  }

  ngOnDestroy() {
    super.ngOnDestroy();
  }
}

@Directive()
export abstract class _BreadcrumbBase extends PaginatedHeaderDirective
  implements AfterContentChecked, AfterContentInit, OnDestroy {
  abstract _items: QueryList<PaginatedHeaderItem & { active: boolean }>;

  @Input()
  get backgroundColor(): ThemePalette {
    return this._backgroundColor;
  }
  set backgroundColor(value: ThemePalette) {
    const classList = this._elementRef.nativeElement.classList;
    classList.remove(`mat-background-${this.backgroundColor}`);

    if (value) {
      classList.add(`mat-background-${value}`);
    }

    this._backgroundColor = value;
  }
  private _backgroundColor: ThemePalette;

  @Input()
  get startWithLast(): boolean | null {
    return this._startWithLast;
  }
  set startWithLast(value: boolean | null) {
    this._startWithLast = coerceBooleanProperty(value);
  }

  @Input()
  get disableRipple() {
    return this._disableRipple;
  }
  set disableRipple(value: any) {
    this._disableRipple = coerceBooleanProperty(value);
  }
  private _disableRipple = false;

  @Input() color: ThemePalette = 'primary';

  private _indexToSelect: number | null = 0;

  constructor(
    elementRef: ElementRef,
    @Optional() dir: Directionality,
    ngZone: NgZone,
    changeDetectorRef: ChangeDetectorRef,
    viewportRuler: ViewportRuler,
    @Optional() platform?: Platform,
    @Optional() @Inject(ANIMATION_MODULE_TYPE) animationMode?: string
  ) {
    super(
      elementRef,
      changeDetectorRef,
      viewportRuler,
      dir,
      ngZone,
      platform,
      animationMode
    );
  }

  protected _itemSelected() {
    // noop
  }

  ngAfterContentInit() {
    this._items.changes
      .pipe(startWith(null), takeUntil(this._destroyed))
      .subscribe(() => {
        this.updateActiveLink();
      });

    super.ngAfterContentInit();
  }

  updateActiveLink(_element?: ElementRef) {
    if (!this._items) {
      return;
    }

    const items = this._items.toArray();

    for (let i = 0; i < items.length; i++) {
      if (items[i].active) {
        this.selectedIndex = i;
        this._changeDetectorRef.markForCheck();
        return;
      }
    }

    this.selectedIndex = -1;
  }
}

@Component({
  selector: 'ucap-breadcrumb',
  exportAs: 'ucapBreadcrumb',
  templateUrl: './breadcrumb.component.html',
  styleUrls: ['./breadcrumb.component.scss'],
  host: {
    class: 'ucap-breadcrumb'
  },
  changeDetection: ChangeDetectionStrategy.OnPush
})
export class BreadcrumbComponent extends _BreadcrumbBase {
  static ngAcceptInputType_disableRipple: BooleanInput;

  @ContentChildren(forwardRef(() => BreadcrumbNodeDirective), {
    descendants: true
  })
  _items: QueryList<BreadcrumbNodeDirective>;

  @ViewChild('breadcrumbListContainer', { static: true })
  _listContainer: ElementRef;

  @ViewChild('breadcrumbList', { static: true })
  _list: ElementRef;

  @ViewChild('nextPaginator')
  _nextPaginator: ElementRef<HTMLElement>;

  @ViewChild('previousPaginator')
  _previousPaginator: ElementRef<HTMLElement>;

  @HostBinding('class.mat-tab-header-pagination-controls-enabled')
  get _paginationControlsEnabled() {
    this.checkSizeChanged();
    return this._showPaginationControls;
  }

  @HostBinding('class.mat-tab-header-rtl')
  get _headerRtl() {
    return this._getLayoutDirection() === 'rtl';
  }

  constructor(
    elementRef: ElementRef,
    @Optional() dir: Directionality,
    ngZone: NgZone,
    changeDetectorRef: ChangeDetectorRef,
    viewportRuler: ViewportRuler,
    /**
     * @deprecated @breaking-change 9.0.0 `platform` parameter to become required.
     */
    @Optional() platform?: Platform,
    @Optional() @Inject(ANIMATION_MODULE_TYPE) animationMode?: string
  ) {
    super(
      elementRef,
      dir,
      ngZone,
      changeDetectorRef,
      viewportRuler,
      platform,
      animationMode
    );
  }
}
