import { merge, of as observableOf, Subject, timer, fromEvent } from 'rxjs';
import { takeUntil } from 'rxjs/operators';

import {
  ChangeDetectorRef,
  ElementRef,
  NgZone,
  Optional,
  QueryList,
  EventEmitter,
  AfterContentChecked,
  AfterContentInit,
  AfterViewInit,
  OnDestroy,
  Directive,
  Inject,
  Input,
  OnInit
} from '@angular/core';
import { ANIMATION_MODULE_TYPE } from '@angular/platform-browser/animations';

import { Direction, Directionality } from '@angular/cdk/bidi';
import { coerceNumberProperty, NumberInput } from '@angular/cdk/coercion';
import { ViewportRuler } from '@angular/cdk/scrolling';
import { FocusKeyManager, FocusableOption } from '@angular/cdk/a11y';
import { END, ENTER, HOME, SPACE, hasModifierKey } from '@angular/cdk/keycodes';
import {
  Platform,
  normalizePassiveListenerOptions
} from '@angular/cdk/platform';

const passiveEventListenerOptions = normalizePassiveListenerOptions({
  passive: true
}) as EventListenerOptions;

export type ScrollDirection = 'after' | 'before';

const EXAGGERATED_OVERSCROLL = 60;

const HEADER_SCROLL_DELAY = 650;

const HEADER_SCROLL_INTERVAL = 100;

export type PaginatedHeaderItem = FocusableOption & {
  elementRef: ElementRef;
};

@Directive()
export abstract class PaginatedHeaderDirective
  implements
    OnInit,
    OnDestroy,
    AfterContentChecked,
    AfterContentInit,
    AfterViewInit {
  static ngAcceptInputType_selectedIndex: NumberInput;

  abstract _items: QueryList<PaginatedHeaderItem>;
  abstract _listContainer: ElementRef<HTMLElement>;
  abstract _list: ElementRef<HTMLElement>;
  abstract _nextPaginator: ElementRef<HTMLElement>;
  abstract _previousPaginator: ElementRef<HTMLElement>;

  private _scrollDistance = 0;

  private _selectedIndexChanged = false;

  protected readonly _destroyed = new Subject<void>();

  _showPaginationControls = false;

  _disableScrollAfter = true;

  _disableScrollBefore = true;

  private _labelCount: number;

  private _scrollDistanceChanged: boolean;

  private _keyManager: FocusKeyManager<PaginatedHeaderItem>;

  private _currentTextContent: string;

  private _stopScrolling = new Subject<void>();

  private _listSize: { width: number; height: number };
  private _listContainerSize: { width: number; height: number };

  @Input()
  disablePagination = false;

  get selectedIndex(): number {
    return this._selectedIndex;
  }
  set selectedIndex(value: number) {
    value = coerceNumberProperty(value);

    if (this._selectedIndex !== value) {
      this._selectedIndexChanged = true;
      this._selectedIndex = value;

      if (this._keyManager) {
        this._keyManager.updateActiveItem(value);
      }
    }
  }
  private _selectedIndex = 0;

  protected _startWithLast = false;

  readonly selectFocusedIndex: EventEmitter<number> = new EventEmitter<
    number
  >();

  readonly indexFocused: EventEmitter<number> = new EventEmitter<number>();

  constructor(
    protected _elementRef: ElementRef<HTMLElement>,
    protected _changeDetectorRef: ChangeDetectorRef,
    private _viewportRuler: ViewportRuler,
    @Optional() private _dir: Directionality,
    private _ngZone: NgZone,
    private _platform?: Platform,
    @Optional() @Inject(ANIMATION_MODULE_TYPE) public _animationMode?: string
  ) {
    _ngZone.runOutsideAngular(() => {
      fromEvent(_elementRef.nativeElement, 'mouseleave')
        .pipe(takeUntil(this._destroyed))
        .subscribe(() => {
          this._stopInterval();
        });
    });

    this._listSize = {
      width: 0,
      height: 0
    };
    this._listContainerSize = {
      width: 0,
      height: 0
    };
  }

  protected abstract _itemSelected(event: KeyboardEvent): void;

  ngOnInit(): void {}

  ngOnDestroy() {
    this._destroyed.next();
    this._destroyed.complete();
    this._stopScrolling.complete();
  }

  ngAfterViewInit() {
    fromEvent(
      this._previousPaginator.nativeElement,
      'touchstart',
      passiveEventListenerOptions
    )
      .pipe(takeUntil(this._destroyed))
      .subscribe(() => {
        this._handlePaginatorPress('before');
      });

    fromEvent(
      this._nextPaginator.nativeElement,
      'touchstart',
      passiveEventListenerOptions
    )
      .pipe(takeUntil(this._destroyed))
      .subscribe(() => {
        this._handlePaginatorPress('after');
      });
  }

  ngAfterContentInit() {
    const dirChange = this._dir ? this._dir.change : observableOf(null);
    const resize = this._viewportRuler.change(150);
    const realign = () => {
      this.updatePagination();
    };
    const windowResize = fromEvent(window, 'resize');

    this._keyManager = new FocusKeyManager<PaginatedHeaderItem>(this._items)
      .withHorizontalOrientation(this._getLayoutDirection())
      .withWrap();

    this._keyManager.updateActiveItem(0);

    typeof requestAnimationFrame !== 'undefined'
      ? requestAnimationFrame(realign)
      : realign();

    merge(dirChange, resize, this._items.changes, windowResize)
      .pipe(takeUntil(this._destroyed))
      .subscribe(() => {
        realign();
        this._keyManager.withHorizontalOrientation(this._getLayoutDirection());
        this.checkSizeChanged();
      });

    this._keyManager.change
      .pipe(takeUntil(this._destroyed))
      .subscribe((newFocusIndex) => {
        this.indexFocused.emit(newFocusIndex);
        this._setFocus(newFocusIndex);
      });
  }

  ngAfterContentChecked(): void {
    if (this._labelCount !== this._items.length) {
      this.updatePagination();
      this._labelCount = this._items.length;
      this._changeDetectorRef.markForCheck();
    }

    if (this._selectedIndexChanged) {
      this._scrollToLabel(this._selectedIndex);
      this._checkScrollingControls();
      this._selectedIndexChanged = false;
      this._changeDetectorRef.markForCheck();
    }

    if (this._scrollDistanceChanged) {
      this._updateScrollPosition();
      this._scrollDistanceChanged = false;
      this._changeDetectorRef.markForCheck();
    }
  }

  _handleKeydown(event: KeyboardEvent) {
    if (hasModifierKey(event)) {
      return;
    }

    switch (event.keyCode) {
      case HOME:
        this._keyManager.setFirstItemActive();
        event.preventDefault();
        break;
      case END:
        this._keyManager.setLastItemActive();
        event.preventDefault();
        break;
      case ENTER:
      case SPACE:
        this.selectFocusedIndex.emit(this.focusIndex);
        this._itemSelected(event);
        break;
      default:
        this._keyManager.onKeydown(event);
    }
  }

  _onContentChanges() {
    const textContent = this._elementRef.nativeElement.textContent;

    if (textContent !== this._currentTextContent) {
      this._currentTextContent = textContent || '';

      this._ngZone.run(() => {
        this.updatePagination();
        this._changeDetectorRef.markForCheck();
      });
    }
  }

  updatePagination() {
    this._checkPaginationEnabled();
    this._checkScrollingControls();
    this._updateScrollPosition();
  }

  checkSizeChanged() {
    const listContainerSize = {
      width: this._listContainer.nativeElement.offsetWidth,
      height: this._listContainer.nativeElement.offsetHeight
    };

    const listSize = {
      width: this._list.nativeElement.offsetWidth,
      height: this._list.nativeElement.offsetHeight
    };

    if (
      this._listSize.width !== listSize.width ||
      this._listSize.height !== listSize.height ||
      this._listContainerSize.width !== listContainerSize.width ||
      this._listContainerSize.height !== listContainerSize.height
    ) {
      this._listContainerSize = listContainerSize;
      this._listSize = listSize;

      this.updatePagination();

      if (this._startWithLast) {
        let nextPaginatorWidth = this._nextPaginator?.nativeElement.offsetWidth;
        let previousPaginatorWidth = this._previousPaginator?.nativeElement
          .offsetWidth;

        const notRendered = !nextPaginatorWidth && !this.disablePagination;

        nextPaginatorWidth = notRendered ? 32 : nextPaginatorWidth;
        previousPaginatorWidth = notRendered ? 32 : previousPaginatorWidth;

        const cWidth = this._listContainer.nativeElement.offsetWidth;
        const lWidth = this._list.nativeElement.scrollWidth;

        if (cWidth === lWidth) {
          this._scrollDistance = 0;
          this._scrollDistanceChanged = true;
          this._checkScrollingControls();
        } else {
          const scrollDistance = Math.max(
            0,
            lWidth - cWidth + (notRendered ? 64 : 0)
          );

          if (0 < scrollDistance) {
            this._scrollDistance = scrollDistance;
            this._scrollDistanceChanged = true;
            this._checkScrollingControls();
          }
        }
      }
    }
  }

  get focusIndex(): number {
    return this._keyManager ? this._keyManager.activeItemIndex! : 0;
  }

  set focusIndex(value: number) {
    if (
      !this._isValidIndex(value) ||
      this.focusIndex === value ||
      !this._keyManager
    ) {
      return;
    }

    this._keyManager.setActiveItem(value);
  }

  _isValidIndex(index: number): boolean {
    if (!this._items) {
      return true;
    }

    const item = this._items ? this._items.toArray()[index] : null;
    return !!item && !item.disabled;
  }

  _setFocus(index: number) {
    if (this._showPaginationControls) {
      this._scrollToLabel(index);
    }

    if (this._items && this._items.length) {
      this._items.toArray()[index].focus();

      const containerEl = this._listContainer.nativeElement;
      const dir = this._getLayoutDirection();

      if (dir === 'ltr') {
        containerEl.scrollLeft = 0;
      } else {
        containerEl.scrollLeft =
          containerEl.scrollWidth - containerEl.offsetWidth;
      }
    }
  }

  _getLayoutDirection(): Direction {
    return this._dir && this._dir.value === 'rtl' ? 'rtl' : 'ltr';
  }

  _updateScrollPosition() {
    if (this.disablePagination) {
      return;
    }

    const scrollDistance = this.scrollDistance;
    const platform = this._platform;
    const translateX =
      this._getLayoutDirection() === 'ltr' ? -scrollDistance : scrollDistance;

    this._list.nativeElement.style.transform = `translateX(${Math.round(
      translateX
    )}px)`;

    if (platform && (platform.TRIDENT || platform.EDGE)) {
      this._listContainer.nativeElement.scrollLeft = 0;
    }
  }

  get scrollDistance(): number {
    return this._scrollDistance;
  }
  set scrollDistance(value: number) {
    this._scrollTo(value);
  }

  _scrollHeader(direction: ScrollDirection) {
    const viewLength = this._listContainer.nativeElement.offsetWidth;

    const scrollAmount = ((direction === 'before' ? -1 : 1) * viewLength) / 3;

    return this._scrollTo(this._scrollDistance + scrollAmount);
  }

  _handlePaginatorClick(direction: ScrollDirection) {
    this._stopInterval();
    this._scrollHeader(direction);
  }

  _scrollToLabel(labelIndex: number) {
    if (this.disablePagination) {
      return;
    }

    const selectedLabel = this._items
      ? this._items.toArray()[labelIndex]
      : null;

    if (!selectedLabel) {
      return;
    }

    const viewLength = this._listContainer.nativeElement.offsetWidth;
    const { offsetLeft, offsetWidth } = selectedLabel.elementRef.nativeElement;

    let labelBeforePos: number;
    let labelAfterPos: number;

    if (this._getLayoutDirection() === 'ltr') {
      labelBeforePos = offsetLeft;
      labelAfterPos = labelBeforePos + offsetWidth;
    } else {
      labelAfterPos = this._list.nativeElement.offsetWidth - offsetLeft;
      labelBeforePos = labelAfterPos - offsetWidth;
    }

    const beforeVisiblePos = this.scrollDistance;
    const afterVisiblePos = this.scrollDistance + viewLength;

    if (labelBeforePos < beforeVisiblePos) {
      this.scrollDistance -=
        beforeVisiblePos - labelBeforePos + EXAGGERATED_OVERSCROLL;
    } else if (labelAfterPos > afterVisiblePos) {
      this.scrollDistance +=
        labelAfterPos - afterVisiblePos + EXAGGERATED_OVERSCROLL;
    }
  }

  _checkPaginationEnabled() {
    if (this.disablePagination) {
      this._showPaginationControls = false;
    } else {
      const isEnabled =
        this._list.nativeElement.scrollWidth >
        this._elementRef.nativeElement.offsetWidth;

      if (!isEnabled) {
        this.scrollDistance = 0;
      }

      if (isEnabled !== this._showPaginationControls) {
        this._changeDetectorRef.markForCheck();
      }

      this._showPaginationControls = isEnabled;
    }
  }

  _checkScrollingControls() {
    if (this.disablePagination) {
      this._disableScrollAfter = this._disableScrollBefore = true;
    } else {
      this._disableScrollBefore = this.scrollDistance === 0;
      this._disableScrollAfter =
        this.scrollDistance === this._getMaxScrollDistance();
      this._changeDetectorRef.markForCheck();
    }
  }

  _getMaxScrollDistance(): number {
    const lengthOfList = this._list.nativeElement.scrollWidth;
    const viewLength = this._listContainer.nativeElement.offsetWidth;

    return lengthOfList - viewLength || 0;
  }

  _stopInterval() {
    this._stopScrolling.next();
  }

  _handlePaginatorPress(direction: ScrollDirection, mouseEvent?: MouseEvent) {
    if (mouseEvent && mouseEvent.button != null && mouseEvent.button !== 0) {
      return;
    }

    this._stopInterval();

    timer(HEADER_SCROLL_DELAY, HEADER_SCROLL_INTERVAL)
      .pipe(takeUntil(merge(this._stopScrolling, this._destroyed)))
      .subscribe(() => {
        const { maxScrollDistance, distance } = this._scrollHeader(direction);

        if (distance === 0 || distance >= maxScrollDistance) {
          this._stopInterval();
        }
      });
  }
  private _scrollTo(position: number) {
    if (this.disablePagination) {
      return { maxScrollDistance: 0, distance: 0 };
    }

    const maxScrollDistance = this._getMaxScrollDistance();
    this._scrollDistance = Math.max(0, Math.min(maxScrollDistance, position));

    this._scrollDistanceChanged = true;
    this._checkScrollingControls();

    return { maxScrollDistance, distance: this._scrollDistance };
  }
}
