import {
  Directive,
  ElementRef,
  EventEmitter,
  Output,
  Input,
  AfterViewInit,
  OnInit,
  OnChanges,
  SimpleChanges
} from '@angular/core';

import { LogService } from '@ucap/ng-logger';

const PATH = 'path';

@Directive({
  selector: 'img[ucapImage]'
})
export class ImageDirective implements OnInit, AfterViewInit, OnChanges {
  @Input()
  base: string;

  @Input()
  path: string;

  @Input()
  validation = true;

  @Input()
  default: string;

  @Output()
  loaded = new EventEmitter<HTMLImageElement>();

  imageSrc: string;

  constructor(
    private elementRef: ElementRef<HTMLImageElement>,
    private logService: LogService
  ) {}

  ngOnInit(): void {
    this.imageSrc = this._validateForPath();
  }

  ngAfterViewInit(): void {
    this.loadImage();
  }

  ngOnChanges(changes: SimpleChanges): void {
    const pathChanges = changes[PATH];

    if (
      !!pathChanges &&
      !pathChanges.firstChange &&
      (pathChanges.previousValue as string) !==
        (pathChanges.currentValue as string)
    ) {
      this.imageSrc = this._validateForPath();
      this.loadImage();
    }
  }

  private loadImage(): void {
    if (this.imageSrc === this.default) {
      this.elementRef.nativeElement.src = this.default;
      this.loaded.emit(this.elementRef.nativeElement);
    } else {
      this.elementRef.nativeElement.src = this.default;

      const imageUrl = `${this.base}${this.path}`;
      const image = new Image();
      image.onload = () => {
        if ('naturalHeight' in image) {
          if (image.naturalHeight + image.naturalWidth === 0) {
            image.onerror('naturalHeight');
            return;
          }
        }
        if (image.width + image.height === 0) {
          image.onerror('width');
          return;
        }
        this.elementRef.nativeElement.src = image.src;
        this.loaded.emit(image);
      };
      image.onerror = (
        event: Event | string,
        source?: string,
        lineno?: number,
        colno?: number,
        error?: Error
      ) => {
        this.logService.error(
          'ucap::ImageDirective',
          event,
          source,
          lineno,
          colno,
          error
        );
        this.elementRef.nativeElement.src = this.default;
        this.loaded.emit(this.elementRef.nativeElement);
      };
      image.src = imageUrl;
    }
  }

  private _validateForPath(): string {
    if (this.validation) {
      if (!this.path || '' === this.path.trim()) {
        return this.default;
      }
    }
    return this.path;
  }
}
