import { Subject, fromEventPattern, Observable } from 'rxjs';
import { takeUntil, share } from 'rxjs/operators';

import {
  Injectable,
  OnDestroy,
  RendererFactory2,
  Renderer2
} from '@angular/core';

@Injectable({ providedIn: 'root' })
export class DomService implements OnDestroy {
  readonly window: Window;

  // tslint:disable-next-line: variable-name
  private readonly _renderer2: Renderer2;

  // tslint:disable-next-line: variable-name
  private readonly _handlerMap: Map<
    any,
    Map<
      string,
      { destroySubject: Subject<void>; $: Observable<Event>; refCount: number }
    >
  >;

  constructor(private rendererFactory2: RendererFactory2) {
    this.window = window;
    this._renderer2 = this.rendererFactory2.createRenderer(null, null);
    this._handlerMap = new Map();
  }

  ngOnDestroy(): void {
    this._handlerMap.forEach((handlers) => {
      handlers.forEach((handler) => {
        if (!!handler.destroySubject) {
          handler.destroySubject.next();
          handler.destroySubject.complete();
        }
      });
      handlers.clear();
    });

    this._handlerMap.clear();
  }

  attach(eventName: string, toTarget: any = this.window): Observable<Event> {
    // tslint:disable-next-line: variable-name
    const _target = !!toTarget ? toTarget : this.window;

    const handler = this._getHandler(eventName, _target);
    if (!!handler) {
      handler.refCount++;
      return handler.$;
    }

    let removeHandler: () => void;
    const addHandler = (callback: (e: Event) => boolean | void) => {
      removeHandler = this._renderer2.listen(_target, eventName, callback);
    };

    const destroySubject: Subject<void> = new Subject();
    const $ = fromEventPattern<Event>(addHandler, () => removeHandler()).pipe(
      takeUntil(destroySubject),
      share()
    );

    this._setHandler(eventName, _target, { destroySubject, $, refCount: 1 });

    return $;
  }

  detach(eventName: string, fromTarget: any = this.window) {
    // tslint:disable-next-line: variable-name
    const _target = !!fromTarget ? fromTarget : this.window;

    const handler = this._getHandler(eventName, _target);
    if (!handler) {
      return;
    }

    handler.refCount--;

    if (0 < handler.refCount) {
      return;
    }

    this._removeHandler(eventName, _target);
  }

  private _getHandler(
    eventName: string,
    fromTarget: any
  ):
    | { destroySubject: Subject<void>; $: Observable<Event>; refCount: number }
    | undefined {
    const eventHandler = this._handlerMap.get(fromTarget);
    if (!eventHandler) {
      return undefined;
    }

    return eventHandler.get(eventName);
  }

  private _setHandler(
    eventName: string,
    toTarget: any,
    handler: {
      destroySubject: Subject<void>;
      $: Observable<Event>;
      refCount: number;
    }
  ) {
    let eventHandler = this._handlerMap.get(toTarget);
    if (!eventHandler) {
      eventHandler = new Map();
      this._handlerMap.set(toTarget, eventHandler);
    }

    if (eventHandler.has(eventName)) {
      throw new Error('Event handler exist already');
    }

    eventHandler.set(eventName, handler);
  }

  private _removeHandler(eventName: string, fromTarget: any) {
    const eventHandler = this._handlerMap.get(fromTarget);
    if (!eventHandler) {
      return;
    }

    const handler = eventHandler.get(eventName);
    if (!handler) {
      return;
    }

    if (!!handler.destroySubject) {
      handler.destroySubject.next();
      handler.destroySubject.complete();
    }
    eventHandler.delete(eventName);

    if (0 === eventHandler.size) {
      this._handlerMap.delete(fromTarget);
    }
  }
}
