import { Subject, fromEvent } from 'rxjs';
import { debounceTime, takeUntil } from 'rxjs/operators';

import {
  Component,
  OnInit,
  Input,
  HostListener,
  Output,
  EventEmitter,
  ViewChild,
  OnDestroy,
  ChangeDetectorRef
} from '@angular/core';
import { WindowState } from '@ucap/native';

@Component({
  selector: 'ucap-title-bar',
  templateUrl: './title-bar.component.html',
  styleUrls: ['./title-bar.component.scss']
})
export class TitleBarComponent implements OnInit, OnDestroy {
  @Input()
  platform: 'darwin' | 'win32' | 'linux' | 'browser';

  @Input()
  windowState: WindowState;

  @Output()
  closed = new EventEmitter<void>();
  @Output()
  minimized = new EventEmitter<void>();
  @Output()
  maximized = new EventEmitter<boolean>();

  @ViewChild('windowsControlsContainer', { static: false })
  windowsControlsContainer: HTMLDivElement;

  isKeyDownAlt = false;
  isMouseOverForMacOsControls = false;
  windowsControlsContainerOffsetHeight: number;

  WindowState = WindowState;

  private ngOnDestroySubject: Subject<void> = new Subject();

  constructor(private changeDetectorRef: ChangeDetectorRef) {}

  ngOnInit() {
    fromEvent(window, 'resize')
      .pipe(takeUntil(this.ngOnDestroySubject), debounceTime(1000))
      .subscribe((event) => {
        if (!!this.windowsControlsContainer) {
          this.windowsControlsContainerOffsetHeight = this.windowsControlsContainer.offsetHeight;
        }
      });
  }

  ngOnDestroy(): void {
    if (!!this.ngOnDestroySubject) {
      this.ngOnDestroySubject.next();
      this.ngOnDestroySubject.complete();
    }
  }

  @HostListener('window:keydown', ['$event'])
  onKeydown(event: KeyboardEvent) {
    if (['AltLeft', 'AltRight'].some((v) => v === event.code)) {
      this.isKeyDownAlt = true;
    } else if (['Escape'].some((v) => v === event.code)) {
      if (WindowState.FullScreen === this.windowState) {
        event.stopPropagation();
        event.preventDefault();
        this.maximized.emit();
      }
    }
  }

  @HostListener('window:keyup', ['$event'])
  onKeyup(event: KeyboardEvent) {
    if (['AltLeft', 'AltRight'].some((v) => v === event.code)) {
      this.isKeyDownAlt = false;
    }
  }

  onClickClose(event: Event) {
    this.closed.emit();
    this.isKeyDownAlt = false;
  }
  onClickMaximize(event: Event) {
    this.maximized.emit(this.isKeyDownAlt);
    this.isKeyDownAlt = false;
  }
  onClickMinimize(event: Event) {
    this.minimized.emit();
    this.isKeyDownAlt = false;
  }

  onMouseenterMacOsControls(event: Event) {
    this.isMouseOverForMacOsControls = true;
  }
  onMouseleaveMacOsControls(event: Event) {
    this.isMouseOverForMacOsControls = false;
  }
}
