import { Observable, Subject } from 'rxjs';
import { distinctUntilChanged, takeUntil } from 'rxjs/operators';

import {
  Directive,
  forwardRef,
  Input,
  OnChanges,
  OnInit,
  OnDestroy,
  Output,
  EventEmitter
} from '@angular/core';

import { coerceNumberProperty, NumberInput } from '@angular/cdk/coercion';
import { ListRange } from '@angular/cdk/collections';

import {
  VirtualScrollStrategy,
  VIRTUAL_SCROLL_STRATEGY,
  VirtualScrollKeyType,
  VirtualScrollToAlign,
  VirtualScrollMeasureSize
} from '../interfaces/virtual-scroll-strategy';
import { VirtualScrollViewportComponent } from '../components/virtual-scroll-viewport';

export interface MeasureSizeMetadata {
  key: VirtualScrollKeyType;
  offset: number;
  size: number;
}

/** Virtual scrolling strategy for lists with items of unknown or dynamic size. */
export class MeasureSizeVirtualScrollStrategy implements VirtualScrollStrategy {
  private _scrolledIndexChange = new Subject<number>();
  /** @docs-private Implemented as part of VirtualScrollStrategy. */
  scrolledIndexChange: Observable<number> = this._scrolledIndexChange.pipe(
    distinctUntilChanged()
  );

  private _measuredSubject: Subject<void> = new Subject();
  measured$: Observable<void> = this._measuredSubject;

  /** The attached viewport. */
  private _viewport: VirtualScrollViewportComponent | null = null;

  /** The minimum amount of buffer rendered beyond the viewport (in pixels). */
  private _minBufferPx: number;

  /** The number of buffer items to render beyond the edge of the viewport (in pixels). */
  private _maxBufferPx: number;

  /** The last measured scroll offset of the viewport. */
  private _lastScrollOffset: number;

  /** The last measured size of the rendered content in the viewport. */
  private _lastRenderedContentSize: number;

  /** The last measured size of the rendered content in the viewport. */
  private _lastRenderedContentOffset: number;

  private _removalFailures = 0;

  /** The estimator used to estimate the size of unseen items. */
  // private _averager: ItemSizeAverager;

  private _metadatas: MeasureSizeMetadata[] = [];
  private _measuredTotalSize = 0;

  /**
   * @param minBufferPx The minimum amount of buffer rendered beyond the viewport (in pixels).
   *     If the amount of buffer dips below this number, more items will be rendered.
   * @param maxBufferPx The number of pixels worth of buffer to shoot for when rendering new items.
   *     If the actual amount turns out to be less it will not necessarily trigger an additional
   *     rendering cycle (as long as the amount of buffer is still greater than `minBufferPx`).
   * @param averager The averager used to estimate the size of unseen items.
   */
  constructor(minBufferPx: number, maxBufferPx: number) {
    this._minBufferPx = minBufferPx;
    this._maxBufferPx = maxBufferPx;
  }

  /**
   * Attaches this scroll strategy to a viewport.
   * @param viewport The viewport to attach this strategy to.
   */
  attach(viewport: VirtualScrollViewportComponent) {
    this._viewport = viewport;
  }

  /** Detaches this scroll strategy from the currently attached viewport. */
  detach() {
    this._scrolledIndexChange.complete();
    this._measuredSubject.complete();
    this._viewport = null;
  }

  useMeasureSize(): boolean {
    return true;
  }

  measureSize(sizes: VirtualScrollMeasureSize[]) {
    if (!sizes || 0 === sizes.length) {
      return;
    }

    const metadatas: MeasureSizeMetadata[] = [];
    let offset = 0;
    sizes.forEach((s, i) => {
      metadatas[i] = { key: s.key, offset, size: s.size };
      offset += s.size;
    });

    console.log('metadatas', metadatas);

    this._measuredTotalSize = offset;
    this._metadatas = metadatas;
    this._updateTotalContentSize();
    this._renderContentForCurrentOffset();

    this._measuredSubject.next();
  }

  measuredSizeOf(key: VirtualScrollKeyType): number {
    const metadata = this._metadatas.find((m) => key === m.key);
    return !!metadata ? metadata.size : -1;
  }

  /**
   * Update the buffer parameters.
   * @param minBufferPx The minimum amount of buffer rendered beyond the viewport (in pixels).
   * @param maxBufferPx The number of buffer items to render beyond the edge of the viewport (in
   *     pixels).
   */
  updateBufferSize(minBufferPx: number, maxBufferPx: number) {
    if (maxBufferPx < minBufferPx) {
      throw 'CDK virtual scroll: maxBufferPx must be greater than or equal to minBufferPx';
    }
    this._minBufferPx = minBufferPx;
    this._maxBufferPx = maxBufferPx;
  }

  /** @docs-private Implemented as part of VirtualScrollStrategy. */
  onContentScrolled() {
    if (this._viewport) {
      this._updateRenderedContentAfterScroll();
    }
  }

  onDataChanged(): void {
    if (this._viewport) {
      // this._checkRenderedContentSize();
      // this._renderContentForCurrentOffset();
    }
  }

  /** @docs-private Implemented as part of VirtualScrollStrategy. */
  onContentRendered() {
    if (this._viewport) {
      this._checkRenderedContentSize();
    }
  }

  /** @docs-private Implemented as part of VirtualScrollStrategy. */
  onRenderedOffsetChanged() {
    if (this._viewport) {
      this._checkRenderedContentOffset();
    }
  }

  /** Scroll to the offset for the given index. */
  scrollToIndex(
    index: number,
    scrollOffset: number,
    align: VirtualScrollToAlign = 'center',
    behavior: ScrollBehavior
  ): void {
    if (this._viewport) {
      const offset = this.offsetForIndex(index, scrollOffset, align);
      this._viewport.scrollToOffset(offset, behavior);
    }
  }

  offsetForIndex(
    index: number,
    scrollOffset: number,
    align: VirtualScrollToAlign = 'smart'
  ): number {
    const viewport = this._viewport!;
    const viewportSize = viewport.getViewportSize();
    const metadata = this._metadatas[index];
    const totalSize = this._measuredTotalSize;

    const maxOffset = Math.max(
      0,
      Math.min(totalSize - viewportSize, metadata.offset)
    );
    const minOffset = Math.max(
      0,
      metadata.offset + metadata.size - viewportSize
    );

    let scrollToAlign = align;

    if ('smart' === scrollToAlign) {
      if (
        scrollOffset >= minOffset - viewportSize &&
        scrollOffset <= maxOffset + viewportSize
      ) {
        scrollToAlign = 'auto';
      } else {
        scrollToAlign = 'center';
      }
    }

    let offset: number;

    switch (scrollToAlign) {
      case 'start':
        offset = maxOffset;
        break;
      case 'end':
        offset = minOffset;
        break;
      case 'center':
        offset = Math.round(minOffset + (maxOffset - minOffset) / 2);
        break;
      case 'auto':
      default:
        if (scrollOffset >= minOffset && scrollOffset <= maxOffset) {
          offset = scrollOffset;
        } else if (scrollOffset < minOffset) {
          offset = minOffset;
        } else {
          offset = maxOffset;
        }
    }

    return offset;
  }

  invalidateMeasurements() {
    this._metadatas = [];
    this._measuredTotalSize = 0;
  }

  private get _averageItemSize() {
    return this._measuredTotalSize / this._metadatas.length;
  }

  /** Update the viewport's total content size. */
  private _updateTotalContentSize() {
    if (!this._viewport) {
      return;
    }
    this._viewport.setTotalContentSize(this._measuredTotalSize);
  }

  private _updateRenderedContentAfterScroll() {
    const viewport = this._viewport!;

    // The current scroll offset.
    const scrollOffset = viewport.measureScrollOffset();
    // The delta between the current scroll offset and the previously recorded scroll offset.
    let scrollDelta = scrollOffset - this._lastScrollOffset; // +: forward, -: backward
    // The magnitude of the scroll delta.
    let scrollMagnitude = Math.abs(scrollDelta);

    const firstVisibleIndex = this._getFirstIndexForOffset(scrollOffset);

    // The currently rendered range.
    const renderedRange = viewport.getRenderedRange();

    let offsetCorrection = 0;
    if (scrollDelta < 0) {
      // The content offset we would expect based on the average item size.
      const predictedOffset = this._metadatas[renderedRange.start].offset;
      // The difference between the predicted size of the unrendered content at the beginning and
      // the actual available space to scroll over. We need to reduce this to zero by the time the
      // user scrolls to the top.
      // - 0 indicates that the predicted size and available space are the same.
      // - A negative number that the predicted size is smaller than the available space.
      // - A positive number indicates the predicted size is larger than the available space
      const offsetDifference =
        predictedOffset - this._lastRenderedContentOffset;
      // The amount of difference to correct during this scroll event. We calculate this as a
      // percentage of the total difference based on the percentage of the distance toward the top
      // that the user scrolled.
      offsetCorrection = Math.round(
        offsetDifference *
          Math.max(
            0,
            Math.min(1, scrollMagnitude / (scrollOffset + scrollMagnitude))
          )
      );

      // Based on the offset correction above, we pretend that the scroll delta was bigger or
      // smaller than it actually was, this way we can start to eliminate the difference.
      scrollDelta = scrollDelta - offsetCorrection;
      scrollMagnitude = Math.abs(scrollDelta);
    }

    // The current amount of buffer past the start of the viewport.
    const startBuffer =
      this._lastScrollOffset - this._lastRenderedContentOffset;
    // The current amount of buffer past the end of the viewport.
    const endBuffer =
      this._lastRenderedContentOffset +
      this._lastRenderedContentSize -
      (this._lastScrollOffset + viewport.getViewportSize());
    // The amount of unfilled space that should be filled on the side the user is scrolling toward
    // in order to safely absorb the scroll delta.
    const underscan =
      scrollMagnitude +
      this._minBufferPx -
      (scrollDelta < 0 ? startBuffer : endBuffer);

    if (underscan > 0) {
      // Check if the scroll magnitude was larger than the viewport size. In this case the user
      // won't notice a discontinuity if we just jump to the new estimated position in the list.
      // However, if the scroll magnitude is smaller than the viewport the user might notice some
      // jitteriness if we just jump to the estimated position. Instead we make sure to scroll by
      // the same number of pixels as the scroll magnitude.
      if (scrollMagnitude >= viewport.getViewportSize()) {
        this._renderContentForCurrentOffset();
      } else {
        // The number of new items to render on the side the user is scrolling towards. Rather than
        // just filling the underscan space, we actually fill enough to have a buffer size of
        // `maxBufferPx`. This gives us a little wiggle room in case our item size estimate is off.
        const addItems = Math.max(
          0,
          Math.ceil(
            (underscan - this._minBufferPx + this._maxBufferPx) /
              this._averageItemSize
          )
        );
        // The amount of filled space beyond what is necessary on the side the user is scrolling
        // away from.
        const overscan =
          (scrollDelta < 0 ? endBuffer : startBuffer) -
          this._minBufferPx +
          scrollMagnitude;
        // The number of currently rendered items to remove on the side the user is scrolling away
        // from. If removal has failed in recent cycles we are less aggressive in how much we try to
        // remove.
        const unboundedRemoveItems = Math.floor(
          overscan / this._averageItemSize / (this._removalFailures + 1)
        );
        const removeItems = Math.min(
          renderedRange.end - renderedRange.start,
          Math.max(0, unboundedRemoveItems)
        );

        // The new range we will tell the viewport to render. We first expand it to include the new
        // items we want rendered, we then contract the opposite side to remove items we no longer
        // want rendered.
        const range = this._expandRange(
          renderedRange,
          scrollDelta < 0 ? addItems : 0,
          scrollDelta > 0 ? addItems : 0
        );
        if (scrollDelta < 0) {
          range.end = Math.max(range.start + 1, range.end - removeItems);
        } else {
          range.start = Math.min(range.end - 1, range.start + removeItems);
        }

        // The new offset we want to set on the rendered content. To determine this we measure the
        // number of pixels we removed and then adjust the offset to the start of the rendered
        // content or to the end of the rendered content accordingly (whichever one doesn't require
        // that the newly added items to be rendered to calculate.)
        let contentOffset: number;
        let contentOffsetTo: 'to-start' | 'to-end';
        if (scrollDelta < 0) {
          const removedSize = viewport.measureRangeSize({
            start: range.end,
            end: renderedRange.end
          });
          // Check that we're not removing too much.
          if (removedSize <= overscan) {
            contentOffset =
              this._lastRenderedContentOffset +
              this._lastRenderedContentSize -
              removedSize;
            this._removalFailures = 0;
          } else {
            // If the removal is more than the overscan can absorb just undo it and record the fact
            // that the removal failed so we can be less aggressive next time.
            range.end = renderedRange.end;
            contentOffset =
              this._lastRenderedContentOffset + this._lastRenderedContentSize;
            this._removalFailures++;
          }
          contentOffsetTo = 'to-end';
        } else {
          const removedSize = viewport.measureRangeSize({
            start: renderedRange.start,
            end: range.start
          });
          // Check that we're not removing too much.
          if (removedSize <= overscan) {
            contentOffset = this._lastRenderedContentOffset + removedSize;
            this._removalFailures = 0;
          } else {
            // If the removal is more than the overscan can absorb just undo it and record the fact
            // that the removal failed so we can be less aggressive next time.
            range.start = renderedRange.start;
            contentOffset = this._lastRenderedContentOffset;
            this._removalFailures++;
          }
          contentOffsetTo = 'to-start';
        }

        // Set the range and offset we calculated above.
        viewport.setRenderedRange(range);
        viewport.setRenderedContentOffset(
          contentOffset + offsetCorrection,
          contentOffsetTo
        );
        this._scrolledIndexChange.next(Math.floor(firstVisibleIndex));
      }
    } else if (offsetCorrection) {
      // Even if the rendered range didn't change, we may still need to adjust the content offset to
      // simulate scrolling slightly slower or faster than the user actually scrolled.
      viewport.setRenderedContentOffset(
        this._lastRenderedContentOffset + offsetCorrection
      );
    }

    // Save the scroll offset to be compared to the new value on the next scroll event.
    this._lastScrollOffset = scrollOffset;
  }

  private _checkRenderedContentSize() {
    const viewport = this._viewport!;
    this._lastRenderedContentSize = viewport.measureRenderedContentSize();
  }

  private _checkRenderedContentOffset() {
    const viewport = this._viewport!;
    this._lastRenderedContentSize = viewport.measureRenderedContentSize();
    this._lastRenderedContentOffset = viewport.getOffsetToRenderedContentStart()!;
  }

  private _renderContentForCurrentOffset() {
    if (!this._viewport) {
      return;
    }
    const viewport = this._viewport!;
    const scrollOffset = viewport.measureScrollOffset();
    this._lastScrollOffset = scrollOffset;

    const firstVisibleIndex = this._getFirstIndexForOffset(scrollOffset);
    const lastVisibleIndex = this._getLastIndexForOffset(
      firstVisibleIndex,
      scrollOffset
    );

    const startBufferIndex = this._findNearestItem(
      Math.max(0, this._metadatas[firstVisibleIndex].offset - this._maxBufferPx)
    );
    const endBufferIndex = this._findNearestItem(
      Math.min(
        this._measuredTotalSize,
        this._metadatas[lastVisibleIndex].offset +
          this._metadatas[lastVisibleIndex].size +
          this._maxBufferPx
      )
    );

    const range = { start: firstVisibleIndex, end: lastVisibleIndex + 1 };
    // const range = this._expandRange(
    //   { start: firstVisibleIndex, end: lastVisibleIndex + 1 },
    //   firstVisibleIndex - startBufferIndex,
    //   endBufferIndex - lastVisibleIndex
    // );

    viewport.setRenderedRange(range);
    viewport.setRenderedContentOffset(
      this._metadatas[firstVisibleIndex].offset
    );
    this._scrolledIndexChange.next(Math.floor(firstVisibleIndex));
  }

  private _expandRange(
    range: ListRange,
    expandStart: number,
    expandEnd: number
  ): ListRange {
    const viewport = this._viewport!;
    const start = Math.max(0, range.start - expandStart);
    const end = Math.min(viewport.getDataLength(), range.end + expandEnd);
    return { start, end };
  }

  private _getFirstIndexForOffset(offset: number) {
    return this._findNearestItem(offset);
  }

  private _getLastIndexForOffset(firstIndex: number, scrollOffset: number) {
    const viewport = this._viewport!;
    const viewportSize = viewport.getViewportSize();
    const itemCount = this._metadatas.length;
    const metadata = this._metadatas[firstIndex];
    const maxOffset = scrollOffset + viewportSize;

    let offset = metadata.offset + metadata.size;
    let lastIndex = firstIndex;

    while (lastIndex < itemCount - 1 && offset < maxOffset) {
      lastIndex++;
      offset += this._metadatas[lastIndex].size;
    }

    return lastIndex;
  }

  private _findNearestItem(offset: number) {
    let high = this._metadatas.length - 1;
    let low = 0;

    while (low <= high) {
      const middle = low + Math.floor((high - low) / 2);
      const currentOffset = this._metadatas[middle].offset;

      if (currentOffset === offset) {
        return middle;
      } else if (currentOffset < offset) {
        low = middle + 1;
      } else if (currentOffset > offset) {
        high = middle - 1;
      }
    }

    if (0 < low) {
      return low - 1;
    } else {
      return 0;
    }
  }
}

/**
 * Provider factory for `AutoSizeVirtualScrollStrategy` that simply extracts the already created
 * `AutoSizeVirtualScrollStrategy` from the given directive.
 * @param autoSizeDir The instance of `CdkAutoSizeVirtualScroll` to extract the
 *     `AutoSizeVirtualScrollStrategy` from.
 */
export function _measureSizeVirtualScrollStrategyFactory(
  measureSizeDir: MeasureSizeVirtualScrollDirective
) {
  return measureSizeDir._scrollStrategy;
}

/** A virtual scroll strategy that supports unknown or dynamic size items. */
@Directive({
  selector: 'ucap-virtual-scroll-viewport[measureSize]',
  providers: [
    {
      provide: VIRTUAL_SCROLL_STRATEGY,
      useFactory: _measureSizeVirtualScrollStrategyFactory,
      deps: [forwardRef(() => MeasureSizeVirtualScrollDirective)]
    }
  ]
})
export class MeasureSizeVirtualScrollDirective
  implements OnInit, OnDestroy, OnChanges {
  /**
   * The minimum amount of buffer rendered beyond the viewport (in pixels).
   * If the amount of buffer dips below this number, more items will be rendered. Defaults to 100px.
   */
  @Input()
  get minBufferPx(): number {
    return this._minBufferPx;
  }
  set minBufferPx(value: number) {
    this._minBufferPx = coerceNumberProperty(value);
  }
  _minBufferPx = 100;

  /**
   * The number of pixels worth of buffer to shoot for when rendering new items.
   * If the actual amount turns out to be less it will not necessarily trigger an additional
   * rendering cycle (as long as the amount of buffer is still greater than `minBufferPx`).
   * Defaults to 200px.
   */
  @Input()
  get maxBufferPx(): number {
    return this._maxBufferPx;
  }
  set maxBufferPx(value: number) {
    this._maxBufferPx = coerceNumberProperty(value);
  }
  _maxBufferPx = 200;

  @Output()
  measured: EventEmitter<void> = new EventEmitter();

  /** The scroll strategy used by this directive. */
  _scrollStrategy = new MeasureSizeVirtualScrollStrategy(
    this.minBufferPx,
    this.maxBufferPx
  );

  private ngOnDestroySubject: Subject<void> = new Subject();

  ngOnInit(): void {
    this._scrollStrategy.measured$
      .pipe(takeUntil(this.ngOnDestroySubject))
      .subscribe(() => {
        this.measured.emit();
      });
  }

  ngOnDestroy(): void {
    if (!!this.ngOnDestroySubject) {
      this.ngOnDestroySubject.next();
      this.ngOnDestroySubject.complete();
    }
  }

  ngOnChanges() {
    this._scrollStrategy.updateBufferSize(this.minBufferPx, this.maxBufferPx);
  }

  invalidateMeasurements() {
    this._scrollStrategy.invalidateMeasurements();
  }

  static ngAcceptInputType_minBufferPx: NumberInput;
  static ngAcceptInputType_maxBufferPx: NumberInput;
}
